from warnings import warn

from pandas import DataFrame

from ..paths import third_party_dir
from ..streams import handle_streams

from ..expression_set import ExpressionSet

from .base import GSEA
from .exceptions import GSEANoResults

#from pathlib import Path
#third_party_dir = Path(third_party_dir)


class cudaGSEA(GSEA):
    path = third_party_dir / 'cudaGSEA/cudaGSEA/src/cudaGSEA'

    def __init__(self, fdr='full', use_cpu=False, **kwargs):
        super().__init__(**kwargs)
        assert self.path.exists()
        self.fdr = fdr
        self.log_tail = []
        self.use_cpu = use_cpu

    def read_from_process(self, process, handlers=None, verbose=False):

        results = []

        def save_result_or_display(text):
            if text.startswith('RESULT: '):
                statistics, gene_set = text[8:].split('\t')
                statistics = statistics.split(' ')
                result = {
                    statistics[i].rstrip(':'): float(statistics[i + 1])
                    for i in range(0, len(statistics), 2)
                }
                assert gene_set.startswith('(') and gene_set.endswith(')')
                result['gene_set'] = gene_set[1:-1]
                results.append(result)
            else:
                self.log_tail.append(text)
                if len(self.log_tail) > 4:
                    self.log_tail = self.log_tail[1:]
                if verbose:
                    self.display(text)

        handlers = {
            'out': save_result_or_display,
            'err': warn
        }

        handle_streams(process, handlers)

        return DataFrame(results)

    metrics_using_variance = {
        'onepass_t_test',
        'onepass_signal2noise',
        'twopass_signal2noise',
        'twopass_t_test',
        'stable_signal2noise',
        'stable_t_test',
        'overkill_signal2noise',
        'overkill_t_test'
    }

    def run(
        self, expression_data: ExpressionSet, gene_sets: str,
        metric='twopass_signal2noise', id_type='symbols',
        permutations=1000, permutation_type='phenotype',
        verbose=False, delete=True, **kwargs
    ):
        super().run(expression_data, gene_sets, metric=metric)

        assert permutation_type == 'phenotype'

        if not gene_sets.endswith('.gmt'):
            gene_sets = self.msigdb.resolve(gene_sets, id_type)

        data_path, classes_path = self.prepare_files(expression_data)

        command = (
            f'{self.path}'
            f' -res {data_path}'
            f' -cls {classes_path}'
            f' -gmx {gene_sets}'
            f' -nperm {permutations}'
            f' -metric {metric}'
            ' -order descending'
            f' -fdr {self.fdr}'
            + (' -cpu' if self.use_cpu else '')
        )
        results = self.run_in_subprocess(command, verbose=verbose)

        if delete:
            self.clean_up(data_path, classes_path)

        cuda_pitfalls_warning = (
            'There might be an error with CUDA (especially after waking the computer up from sleep); '
            'if you run your graphics server (X/Wayland) on a separate card, '
            'you can try using: sudo rmmod nvidia_uvm; sudo modprobe nvidia_uvm'
        )

        if results.empty:
            last_line = self.log_tail[-1]
            if 'error' in last_line.lower():
                if not verbose:
                    print('Command:', command)
                if 'Error: duplicate symbol' in last_line:
                    print(
                        'Duplicate symbol error may be caused by duplicated genes in index, or '
                        'by having an empty column in the data. Please inspect this data frame: '
                    )
                    print(expression_data.joined)
                if 'CUDA error' in last_line:
                    warn(cuda_pitfalls_warning)
            print('Tail of the logs:')
            print(self.log_tail)
            raise GSEANoResults('No results returned.')

        results = results.set_index('gene_set').rename({
            'ES': 'es',
            'FWER': 'fwer_p-val',
            'NES': 'nes',
            'NP': 'nom_p-val',
            'FDR': 'fdr_q-val'
        }, axis=1)

        return {
            expression_data.case_name: results[results.nes > 0],
            expression_data.control_name: results[results.nes < 0]
        }



import re
from glob import glob
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory, NamedTemporaryFile
from typing import Union
from warnings import warn

from pandas import read_table, DataFrame

from .base import GSEA
from ..paths import tmp_dir, third_party_dir


class GSEADesktop(GSEA):
    path = Path('java')

    def __init__(self, memory_size=5000, gsea_jar_path: Union[str, Path] = None, **kwargs):
        super().__init__(**kwargs)
        self.memory_size = memory_size
        if gsea_jar_path is None:
            gsea_jar_path = third_party_dir / 'gsea-3.0.jar'
        else:
            gsea_jar_path = Path(gsea_jar_path)
            if not str(gsea_jar_path).endswith('.jar'):
                warn(f'{gsea_jar_path} does not look like a valid GSEADesktop path')
        self.gsea_path = gsea_jar_path
        if not self.gsea_path.exists():
            raise Exception(
                f'Could not find GSEADesktop installation in {gsea_jar_path};'
                f' please symlink it or set path in "gsea_jar_path" argument.'
            )

    @property
    def core_command(self):
        return f'{self.path} -cp {self.gsea_path}'

    @staticmethod
    def load_results(out_dir, name, expression_data, delete=True):
        timestamps = []
        for path in glob(f'{out_dir}/{name}.Gsea.*'):
            match = re.match(rf'{name}\.Gsea\.(\d*)', Path(path).name)
            timestamp = match.group(1)
            timestamps.append(timestamp)

        timestamps = sorted(timestamps, reverse=True)
        newest = timestamps[0]

        results = {}
        for class_type in set(expression_data.safe_classes):
            path = f'{out_dir}/{name}.Gsea.{newest}/gsea_report_for_{class_type}_{newest}.xls'
            result: DataFrame = read_table(path)
            assert all(result['Unnamed: 11'].isnull())
            result.drop('Unnamed: 11', axis='columns', inplace=True)
            result.columns = [column.lower().replace(' ', '_') for column in result.columns]
            result.drop('gs<br>_follow_link_to_msigdb', axis='columns', inplace=True)
            result = result.sort_values('fdr_q-val')
            result.set_index('name', inplace=True)

            results[class_type] = result.dropna(subset=['nes', 'fdr_q-val'])

        if delete:
            rmtree(f'{out_dir}/{name}.Gsea.{newest}', ignore_errors=True)

        return results

    metrics_using_variance = {'tTest', 'Signal2Noise'}

    def run(
        # Common GSEA interface
        self, expression_data, gene_sets, metric='Signal2Noise', id_type='symbols',
        permutations=1000, permutation_type='phenotype', verbose=False,
        # GSEADesktop-specific
        collapse=True, mode='Max_probe', detailed_sets_analysis=False, out_dir=None,
        plot_top_x=0, name='my_analysis', normalization='meandiv', min_genes=15, max_genes=500,
        sort='real',
        # internal, won't be passed to GSEA
        use_existing=False, delete=True,
        # additional arguments to be passed
        **kwargs
    ):
        """Memory in MB"""
        super().run(expression_data, gene_sets, metric=metric)

        if isinstance(gene_sets, str):
            assert id_type in {'symbols', 'entrez'}

            if not gene_sets.endswith('.gmt'):
                gene_sets = self.msigdb.resolve(gene_sets, id_type)
        else:
            with NamedTemporaryFile(mode='w', delete=False, suffix='.gmt') as f:
                gene_sets.to_gmt(f)
                gene_sets = f.name

        assert permutation_type in {'Gene_set', 'phenotype'}
        assert metric in {'Diff_of_Classes', 'log2_Ratio_of_Classes', 'Ratio_of_Classes' 'tTest', 'Signal2Noise'}

        if use_existing:
            try:
                results = self.load_results(out_dir, name, expression_data)
                print(
                    'Re-using pre-calculated results for GSEA'
                    ' (to prevent this, set use_existing=False or clear the temporary GSEA folder)'
                )
                return results
            except IndexError:
                pass

        command = f'{self.core_command} -Xmx{self.memory_size}m xtools.gsea.Gsea'

        for key, value in kwargs.items():
            command += f' -{key} {value}'

        data_path, classes_path = self.prepare_files(expression_data)

        with TemporaryDirectory(dir=tmp_dir) as temp_dir:
            out_dir = self.prepare_out_dir(out_dir, temp_dir)

            command += (
                f' -res {data_path}'
                f' -cls {classes_path}'
                f' -gmx {gene_sets}'
                f' -collapse false -mode {mode} -norm {normalization}'
                f' -nperm {permutations} -permute {permutation_type} -rnd_type no_balance'
                f' -scoring_scheme weighted -rpt_label {name}'
                f' -metric {metric}'
                f' -sort {sort} -order descending'
                ' -create_gcts false -create_svgs false'
                # ' -include_only_symbols true'
                f' -make_sets {str(detailed_sets_analysis).lower()} -median false -num 100'
                f' -plot_top_x {plot_top_x} -rnd_seed timestamp'
                ' -save_rnd_lists false'
                f' -set_max {max_genes} -set_min {min_genes}'
                ' -zip_report false'
                f' -out {out_dir}'
                ' -gui false'
            )
            self.run_in_subprocess(command, verbose=verbose)

            results = self.load_results(out_dir, name, expression_data, delete=delete)

        if delete:
            self.clean_up(data_path, classes_path)

        return results



import re
from collections import Counter, defaultdict
from functools import lru_cache
from glob import glob
from pathlib import Path
from typing import Collection, Dict, List, Set
from warnings import warn

from pandas import DataFrame


class GeneSet:

    def __init__(self, name: str, genes: Collection[str], description: str = None):
        self.name = name
        self.genes = frozenset(genes)
        self.description = description

        if len(genes) == 0:
            warn(f'GeneSet {repr(name)} is empty')

        redundant_genes = None

        if len(genes) != len(self.genes):
            redundant_genes = {gene: count for gene, count in Counter(genes).items() if count > 1}

            warn(f'GeneSet {repr(name)} received a non-unique collection of genes; redundant genes: {redundant_genes}')

        self.redundant_genes = redundant_genes

    @classmethod
    def from_gmt_line(cls, line):
        name, description, *ids = line.strip().split('\t')
        return cls(name, ids, description)

    def __repr__(self):
        genes = ': ' + (', '.join(sorted(self.genes))) if len(self.genes) < 5 else ''
        return f'<GeneSet {repr(self.name)} with {len(self.genes)} genes{genes}>'


class GeneSets:

    def __init__(self, gene_sets: Collection[GeneSet], name='', allow_redundant=False):
        self.gene_sets = tuple(gene_sets)
        self.name = name
        if not allow_redundant:
            redundant = self.find_redundant()
            if any(redundant):
                message = 'Provided gene sets are not redundant; '
                if len(redundant) > 3:
                    message += (
                        f'there are {len(redundant)} gene sets having more than one name assigned; '
                        'use find_redundant() to investigate further.'
                    )
                else:
                    identical = ', '.join(
                        ' and '.join(map(repr, pathways)) + f' ({len(gene_set)} genes)'
                        for gene_set, pathways in redundant.items()
                    )
                    message += f'following gene sets are identical: {identical}'
                warn(message)

    def group_identical(self, key='name') -> Dict[frozenset, List[str]]:
        pathways_by_gene_set = defaultdict(list)
        for pathway in self.gene_sets:
            pathways_by_gene_set[pathway.genes].append(getattr(pathway, key))
        return pathways_by_gene_set

    def find_redundant(self, key='name', min_duplicates=1) -> Dict[frozenset, List[str]]:
        return {
            gene_set: pathways
            for gene_set, pathways in self.group_identical(key=key).items()
            if len(pathways) > min_duplicates
        }

    @classmethod
    def from_gmt(cls, path, name=None):
        with open(path) as f:
            return cls({
                GeneSet.from_gmt_line(line)
                for line in f
            }, name=name or Path(path).name)

    def trim(self, min_genes, max_genes: int):
        return GeneSets({
            gene_set
            for gene_set in self.gene_sets
            if min_genes <= len(gene_set.genes) <= max_genes
        })

    def format_names(self, formatter):
        for gene_set in self.gene_sets:
            gene_set.name = formatter(gene_set.name)
        return self

    def _to_gmt(self, f):
        for gene_set in self.gene_sets:
            f.write(gene_set.name + '\t' + '\t'.join(gene_set.genes) + '\n')

    def to_gmt(self, path):
        if isinstance(path, str):
            with open(path, mode='w') as f:
                self._to_gmt(f)
        else:
            self._to_gmt(path)

    def subset(self, genes: Set[str]):
        return GeneSets({
            GeneSet(name=gene_set.name, genes=gene_set.genes & genes)
            for gene_set in self.gene_sets
        })

    @property
    @lru_cache()
    def all_genes(self):
        all_genes = set()

        for gene_set in self.gene_sets:
            all_genes.update(gene_set.genes)

        return all_genes

    def to_frame(self) -> DataFrame:
        all_genes = self.all_genes
        return DataFrame(
            [
                [
                    gene in gene_set.genes
                    for gene in all_genes
                ]
                for gene_set in self.gene_sets
            ],
            index=[gene_set.name for gene_set in self.gene_sets],
            columns=all_genes
        )

    @property
    @lru_cache()
    def gene_sets_by_name(self):
        by_name = {
            gene_set.name: gene_set
            for gene_set in self.gene_sets
        }
        assert len(self.gene_sets) == len(by_name)
        return by_name

    def __len__(self):
        return len(self.gene_sets)

    def __repr__(self):
        name = ' ' + repr(self.name) if self.name else ''
        return f'<GeneSets{name} with {len(self.gene_sets)} gene sets>'


# for backwards compatibility
GeneMatrixTransposed = GeneSets


class MolecularSignaturesDatabase:
    def __init__(self, path, version='6.2'):
        self.path = Path(path)
        self.version = version
        wildcard_path = str((self.path / f'*.v{self.version}.*.gmt').resolve())
        self.gene_sets = [
            self.parse_name(Path(p).name)
            for p in glob(wildcard_path)
        ]

    def parse_name(self, name):
        parsed = re.match(rf'(?P<name>.*?)\.v{self.version}\.(?P<id_type>(entrez|symbols)).gmt', name)
        return parsed.groupdict()

    def resolve(self, gene_sets, id_type):
        path = self.path / f'{gene_sets}.v6.2.{id_type}.gmt'
        if path.exists():
            return str(path)
        else:
            raise ValueError(f'Unknown library: {path}!')

    def load(self, gene_sets, id_type) -> GeneSets:
        path = self.resolve(gene_sets=gene_sets, id_type=id_type)

        return GeneSets.from_gmt(path, name=gene_sets)



from copy import copy
from os import makedirs
from os import remove
from pathlib import Path
from subprocess import Popen, PIPE
from tempfile import NamedTemporaryFile
from typing import Set
from warnings import warn

from pandas import DataFrame
from jupyter_helpers.following_tail import FollowingTail

from abc import ABC, abstractmethod, abstractproperty

from ..streams import handle_streams
from ..expression_set import ExpressionSet

from .exceptions import GSEANotEnoughSamples
from ..paths import tmp_dir


class GSEA(ABC):

    def __init__(self, display=None, display_error=warn):
        self._original_display = display
        self.display = self.get_display(display)
        self.display_error = display_error
        #self.msigdb = MolecularSignaturesDatabase()

    @staticmethod
    def get_display(display):
        if not display:
            display = FollowingTail()
        return display

    def prepare_output(self):
        self.display = self.get_display(self._original_display)
        if isinstance(self.display, FollowingTail):
            self.display.activate()

    @staticmethod
    def prepare_files(expression_data: ExpressionSet):
        with NamedTemporaryFile('w', delete=False, dir=tmp_dir, suffix='.cls') as f:
            classes_path = f.name
            expression_data.to_cls(f)

        with NamedTemporaryFile('w', delete=False, dir=tmp_dir, suffix='.gct') as f:
            expression_data.to_gct(f)
            data_path = f.name

        return data_path, classes_path

    @property
    def default_handlers(self):
        return {
            'out': self.display,
            'err': self.display_error
        }

    def read_from_process(self, process, handlers=None, verbose=False):
        if handlers is None:
            handlers = copy(self.default_handlers)

        if not verbose:
            del self.default_handlers['out']

        return handle_streams(process, handlers)

    def run_in_subprocess(self, command, verbose=False):
        if verbose:
            print(command)
        arguments = command.split(' ')
        process = Popen(arguments, stdout=PIPE, stderr=PIPE)
        return self.read_from_process(process, verbose=verbose)

    @staticmethod
    def prepare_out_dir(out_dir, temp_dir):
        if not out_dir:
            out_dir = Path(temp_dir)
        else:
            out_dir = Path(out_dir)
            if not out_dir.is_absolute():
                pwd = Path().absolute().parent
                out_dir = pwd / str(out_dir)
            makedirs(out_dir, exist_ok=True)
        return out_dir

    @abstractproperty
    def metrics_using_variance(self) -> Set[str]:
        """Metrics which would fail if given less than three samples per category"""

    def has_enough_samples_for_metric(self, expression_data, metric: str):
        if metric in self.metrics_using_variance:
            # TODO
            return True
            #return len(expression_data.controls.columns) >= 3 and len(expression_data.cases.columns) >= 3
        return True

    @abstractmethod
    def run(self, expression_data: ExpressionSet, gene_sets: str, metric: str = None, **kwargs) -> DataFrame:
        """Run the GSEA method using case and control data from given expression object.

        Raises:
            GSEANoResults exception if the method fails with no results.
            GSEANotEnoughSamples if specified metric requires more samples than provided.
        """

        if not self.has_enough_samples_for_metric(expression_data, metric):
            warn(f'Too few samples for the metric {metric}')
            raise GSEANotEnoughSamples(f'Too few samples for the metric {metric}')

    @staticmethod
    def clean_up(data_path, classes_path):
        remove(data_path)
        remove(classes_path)



from pathlib import Path
from tempfile import TemporaryDirectory

from pandas import read_csv

from ..paths import third_party_dir

from .base import GSEA


class GSEApy(GSEA):
    """
    To use gsea.py please install it with:
        >>> pip3 install gseapy
    and link its binary to the thirdparty directory
        >>> ln -s virtual_environment_path/bin/gseapy thirdparty/gseapy
    """

    path = Path(third_party_dir / 'gseapy')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if not self.path.exists():
            print(self.__doc__)

        import gseapy
        self.libraries = gseapy.get_library_name()

    metrics_using_variance = {
        'signal_to_noise',
        't_test'
    }

    def run(
        self, expression_data, gene_sets='KEGG_2016', metric='signal_to_noise', id_type='symbols',
        permutations=1000, permutation_type='phenotype', verbose=True,
        program='gsea', processes=8, plot=False, out_dir=None, **kwargs
    ):

        super().run(expression_data, gene_sets, metric=metric)

        if not gene_sets.endswith('.gmt'):
            if gene_sets in self.libraries:
                assert id_type == 'symbols'
            else:
                gene_sets = self.msigdb.resolve(gene_sets, id_type)

        with TemporaryDirectory() as temp_dir:
            out_dir = self.prepare_out_dir(out_dir, temp_dir)

            data_path, classes_path = self.prepare_files(expression_data)

            kwargs.update({
                'data': data_path,
                'cls': classes_path,
                'permu-type': permutation_type,
                'permu-num': permutations,
                'threads': processes,
                'gmt': gene_sets,
                'method': metric,
                'outdir': str(out_dir)
            })

            command = ' '.join([
                str(self.path.expanduser()),
                program,
                *[
                    f'--{name} {value}'
                    for name, value in kwargs.items()
                ]
            ])
            command += ' --verbose' if verbose else ''
            command += ' --no-plot' if not plot else ''

            self.run_in_subprocess(command, verbose=verbose)

            result = read_csv(out_dir / f'gseapy.gsea.{permutation_type}.report.csv')
            return result



from select import select
from types import FunctionType
from typing import Dict


def handle_streams(process, handlers: Dict[str, FunctionType]):
    stream_to_handler = {
        stream: handlers[stream_name]
        for stream, stream_name in [
            (process.stdout, 'out'),
            (process.stderr, 'err')
        ]
        if stream_name in handlers
    }

    descriptor_to_stream = {
        stream.fileno(): stream
        for stream in stream_to_handler
    }

    active_descriptors = list(descriptor_to_stream)

    while True:
        selected_descriptors, _, _ = select(active_descriptors, [], [])

        empty_output = True

        for descriptor in selected_descriptors:
            stream = descriptor_to_stream[descriptor]
            line = stream.readline()
            if line:
                line = line.decode('utf-8')
                handler = stream_to_handler[stream]
                handler(line.rstrip())
                empty_output = False

        if empty_output and process.poll() is not None:
            return



# from helpers.temp import create_tmp_dir, mount_root_in_ramdisk
from pathlib import Path

tmp = '/tmp/'

tmp_dir = tmp + 'gsea/input'
gsea_home = tmp + 'gsea_home'

Path(tmp_dir).mkdir(parents=True, exist_ok=True)
Path(gsea_home).mkdir(parents=True, exist_ok=True)

# tmp_dir = create_tmp_dir('gsea/input')
# gsea_home = mount_root_in_ramdisk('~/gsea_home')
third_party_dir = Path(__file__).parent / 'thirdparty'



import pandas
from numpy import savetxt
from pandas import DataFrame, Series, concat
from typing.io import TextIO
from typing import List, Union


class ExpressionSet:
    # TODO: have a limit() method (to trim the data to the most relevant genes only)

    joined: DataFrame
    classes: Series

    def __init__(self, data: DataFrame, classes: Union[Series, List]):
        self.joined = data
        self.classes = Series(classes)
        if len(data.columns) != len(classes):
            raise ValueError('Number of classes different from the number of columns')

    @classmethod
    def from_cases_and_controls(
        cls, cases: DataFrame, controls: DataFrame,
        case_name: str = 'case', control_name: str = 'control'
    ) -> 'ExpressionSet':
        assert cases.index == controls.index
        return cls(
            data=concat(cases, controls),
            classes=Series(
                [case_name] * len(cases.columns) + [control_name] * len(controls.columns)
            )
        )

    #def __repr__(self):
    #    return f'<{self.__class__.__name__}: {len(self.cases.columns)} cases, {len(self.controls.columns)} controls>'

    def contrast(self, case, control) -> 'ExpressionSet':
        """Select a subset of two groups (cases and controls) from experimental groups in the expression set.

        For example, having groups Drug_A, Drug_B, Control, if you want to only analyse Drug_A vs Control, specify:

        subset = expression_set.contrast(case='Drug_A', control='Control')
        """
        assert case in self.classes.unique() and control in self.classes.unique()
        mask = self.classes.isin({case, control})
        return ExpressionSet(
            self.joined[self.joined.columns[mask]],
            self.classes[mask]
        )

    @property
    def hashable(self):
        all_data = self.joined
        return tuple(all_data.index), tuple(all_data.columns), all_data.sum().sum()

    @property
    def safe_classes(self):
        return [cls.replace(' ', '_') for cls in self.classes]

    def to_cls(self, f):
        classes = self.safe_classes
        f.write(f'{len(classes)} {len(set(classes))} 1\n')
        classes_set = []
        for class_ in classes:
            if class_ not in classes_set:
                classes_set.append(class_)
        f.write(f'# {" ".join(classes_set)}\n')
        f.write(' '.join(classes))

    def to_gct(self, f: TextIO, tabular_writer='to_txt'):
        f.write('#1.2\n')
        expression_data = self.joined
        assert expression_data.notnull().all().all()
        f.write(f'{len(expression_data)}\t{len(expression_data.columns)}\n')
        getattr(self, tabular_writer)(f, expression_data)

    def to_txt(self, f: TextIO, expression_data=None):
        if expression_data is None:
            from copy import copy
            expression_data = copy(self.joined)
        columns = expression_data.columns
        pandas.options.mode.chained_assignment = None
        expression_data['Description'] = 'na'
        expression_data = expression_data[['Description', *columns]]
        if type(expression_data.index[0]) is bytes:
            expression_data.index = [b.decode('utf-8') for b in expression_data.index]
        expression_data.index = expression_data.index.astype(str)
        expression_data.index.name = 'gene'
        header = '\t'.join([expression_data.index.name, *expression_data.columns]) + '\n'
        f.write(header)

        savetxt(
            f,
            expression_data.reset_index().values,
            delimiter='\t',
            # entrez id or symbol (as str), 'na' (not a description, str), *data (floats)
            fmt='%s\t%s' + ('\t%f' * (len(expression_data.columns) - 1))
        )
